/*
 * Copyright (C) 2008-2024 OpenSIPS Solutions
 * Copyright (C) 2004-2006 Voice Sistem SRL
 *
 * This file is part of Open SIP Server (opensips).
 *
 * opensips is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * opensips is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,USA
 *
 */


#ifndef _SQLOPS_PARSE_H_
#define _SQLOPS_PARSE_H_

#include "../../str.h"
#include "../../usr_avp.h"
#include "sqlops_impl.h"
#include "sqlops_db.h"


int parse_avp_db(char *s, struct db_param *dbp, int allow_scheme);

int parse_avp_db_scheme(char *s, struct db_scheme *scheme);

#endif

